--[[

This file is part of RampageAI.

RampageAI is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

RampageAI is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with RampageAI; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

--]]

-------------------------------
-- Initialization Supporters --
-------------------------------

-- Function to capture the original API functions and replace them
function InitAPIFuncs()
	-- RAI_Require is overridden before the script "require" commands
	-- RAI_TraceAI is overridden with debug initialization code
	RAI_MoveToOwner = MoveToOwner
	RAI_Move = Move
	RAI_Attack = Attack
	-- RAI_GetV = GetV			-- inefficient to override
	-- RAI_GetActors = GetActors		-- inefficient to override
	-- RAI_GetTick = GetTick		-- unwise to override
	-- RAI_GetMsg = GetMsg			-- no need to override
	-- RAI_GetResMsg = GetResMsg		-- no need to override
	RAI_SkillObject = SkillObject
	-- RAI_SkillGround = SkillGround	-- disabled by RO
	RAI_IsMonster = IsMonster

	-- Replace action functions with ones that check portal distance
	MoveToOwner = function(id)
		if not EnteringPortal then
			return RAI_MoveToOwner(id)
		end
	end
	Move = function(id,x,y)
		if not EnteringPortal then
			return RAI_Move(id,x,y)
		end
	end
	Attack = function(id1,id2)
		if not EnteringPortal then
			return RAI_Attack(id1,id2)
		end
	end
	SkillObject = function(id,level,skill,target)
		if not EnteringPortal then
			return RAI_SkillObject(id,level,skill,target)
		end
	end

	-- Set IsMonster to respond according to WoE-zone status
	--	* idea from Fallen~Angel of RO-Empire forums *
	IsMonster = function(id)
		-- Check if we're in a WoE zone
		if WoEZone then
			-- Log it
			Log(0,"Using WoE-customized IsMonster() function.")

			-- Setup a new function to determine who is an enemy and who isn't
			IsMonster = function(actor)
				-- Check if regular function returns 1
				if RAI_IsMonster(actor) == 1 then
					return 1
				end

				-- Check if the actor is in our data
				if ActorData[actor] == nil then
					-- Don't know anything about them, so return 0
					return 0
				end

				-- Check if they're in our permanent friend list
				if IsPermFriend(actor) then
					return 0
				end

				-- Otherwise they're not a friend
				return 1
			end
		else
			-- Use normal IsMonster
			IsMonster = RAI_IsMonster
		end

		-- Whichever IsMonster function we're using, return its value
		return IsMonster(id)
	end

	-- Never capture these again
	InitAPIFuncs = NoReturn
end

-- Function to determine the way we should acquire targets
function InitAcquire()
	-- Default to allowing the same targets
	AcquireSupport_HomunsTarget = FalseReturn
	AcquireSupport_AlchsTarget = FalseReturn

	-- Decide how to acquire targets while in passive mode
	if PassiveDefense then
		PassiveAcquire	= AI_PassiveAcquire	-- defensive only
	else
		-- Check if still want to assist
		if not AssistOwner then
			PassiveAcquire = NegativeReturn	-- never acquire a target
		else
							-- assist, but don't defend
			PassiveAcquire = AI_AssistOwnerAcquire
		end
	end

	-- Aggro automatically acquires all valid targets
	AggroAcquire = AI_AggroAcquire

	-- Decide how we want to target our alchemist abilities
	if not AlchTargetSeparate then
		-- Alchemist will always acquire to support the homunculus
		AlchAggroAcquire = AI_AlchSupportAcquire
		AlchPassiveAcquire = AI_AlchSupportAcquire
	else
		AlchAggroAcquire = AI_AggroAcquire
		if PassiveDefense then
			AlchPassiveAcquire = AI_PassiveAcquire
		else
			AlchPassiveAcquire = AI_AlchSupportAcquire
		end

		-- Check if they want to force different targets
		if AlchTargetForceDifferent then
			-- Don't allow the same targets
			AcquireSupport_HomunsTarget = AI_AcquireSupport_HomunsTarget
			AcquireSupport_AlchsTarget = AI_AcquireSupport_AlchsTarget
		end
	end

	-- Check if we should try to assist our owner
	if AssistOwner then
		AssistAcquire = AI_AssistOwnerAcquire	-- assist our owner when possible
	else
		AssistAcquire = NegativeReturn		-- never assist
	end

	-- Check if we need to modify values for prioritizing defense
	if PassiveDefense then
		-- If owner defense is prioritized with 0 or fewer
		-- monsters, set it to 1000 or more monsters
		-- (essentially disabled)
		if PrioritizeOwnerDefense < 1 then
			PrioritizeOwnerDefense = 1000
		end

		-- Same thing for reprioritization of the homunculus
		if PrioritizeHomunDefense < 1 then
			PrioritizeHomunDefense = 1000
		end

		-- Check the friend defense
		if PrioritizeFriendDefense < 1 then
			PrioritizeFriendDefense = 1000
		end
	end
end

-- Function to initialize attack log maintaining
function InitAttackLog()
	-- Check if its enabled
	if EnableAttacksTimeout then
		-- Enabled
		TrimAttackLog = AI_TrimAttackLog
		ResetAttackLog = AI_ResetAttackLog
		AddAttackLog = AI_AddAttackLog
	else
		-- Not enabled
		TrimAttackLog = ZeroZeroReturn
		ResetAttackLog = NoReturn
		AddAttackLog = NoReturn
	end
end

-- Function to initialize kill-steal checking
function InitKillSteal()
	-- Do we care about kill-stealing?
	if AllowKS then
		-- Nope
		KillSteal = FalseReturn
	else
		-- Yep
		KillSteal = AI_KillSteal
	end
end

-- Function to initialize HP delta gathering
function InitHPDeltas()
	-- Check if we need the alchemist's HP delta
	if false then
		-- Not used yet
		CalculateHPDeltasAlch = AI_CalculateHPDeltasAlch
	else
		-- We don't want to waste time on this
		CalculateHPDeltasAlch = NoReturn
	end

	-- Check if we need the homunculus's HP delta
	if AutoRes or (AutoAidPot and AidPotAdvanced) then
		-- We use this for advanced aid potioning, and for auto-res
		CalculateHPDeltasHomu = AI_CalculateHPDeltasHomu
	else
		-- We don't want to waste time on this
		CalculateHPDeltasHomu = NoReturn
	end
end

-- Function to initialize manual skill usage behaviour
function InitManualSkills()
	-- Check if skills should cause attacks
	if SkillsCauseAttack then
		-- Set the target after manually using a skill
		CMD_SkillObj = CMD_SkillObjAndAttack
	else
		-- Don't set the target after manually using a skill
		CMD_SkillObj = CMD_SkillObj_Original
	end
end

-- Function to initialize percent/amount functions
function InitAmountOrPercent()
	-- Check if we want to use percentages for SP
	if PercentSP then
		-- Use functions with percentage versions
		SPUsable = SPUsablePercent
		ActorSP = ActorSPPercent
	else
		-- Use the normal amount versions
		SPUsable = SPUsableAmount
		ActorSP = ActorSPAmount
	end

	-- Check if we want to use percentages for HP
	if PercentHP then
		-- Use functions with percentage versions
		ActorHP = ActorHPPercent
	else
		-- Use the normal amount versions
		ActorHP = ActorHPAmount
	end
end


-- Function to initialize automatic usage of homunculus skills
function InitHomunSkills()
	-- Check if skills are allowed
	if AllowSkills then
		-- Setup to allow skills

		-- Check if we have any attack skills
		if Table.getn(AttackSkills) < 1 and
			Table.getn(RangedSkills) < 1 then
			
			-- Log it
			Log(0,"Disabling all attack skills, for lack any.")

			-- We don't, so don't bother trying to use any
			SendAttack = AI_Attack
			AttackSkillReady = FalseReturn
			UseAutoSkill = NoReturn
			CheckAttackSkill = ZeroReturn
			MSkillsAllowed = FalseReturn
		else
			-- We do have some attack skills, so we'll try to use 'em
			SendAttack = AI_SkilledAttack
			UseAutoSkill = AI_UseSkill
			AttackSkillReady = AI_AttackSkillReady
			MSkillsAllowed = AI_MSkillsAllowed

			-- Check if we're ignoring skill ranges
			if not IgnoreAttackSkillRange then
				CheckAttackSkill = AI_CheckAttackSkill
			else
				CheckAttackSkill = AI_CheckAttackSkill_NoRange
			end
		end

		-- Even if we don't have attack skills, we still have others
		SkillAllowed = AI_SkillAllowed
	else
		-- Skills aren't allowed
		SendAttack = AI_Attack
		CheckAttackSkill = ZeroReturn
		UseAutoSkill = NoReturn
		AttackSkillReady = FalseReturn
		SkillAllowed = FalseReturn
		MSkillsAllowed = FalseReturn
	end
end
