--[[

This file is part of RampageAI.

RampageAI is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

RampageAI is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with RampageAI; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

--]]

------------------------
-- Distance Functions --
------------------------

-- Position->Position Pythagorean Distance
function PPDistance(x1,y1,x2,y2)
	-- Check all our parameters
	if type(x1) ~= "number" or
		type(y1) ~= "number" or
		type(x2) ~= "number" or
		type(y2) ~= "number"
	then
		-- Something wasn't a number, return nil so its easier to trace
		return nil
	end

	--return math.floor(math.sqrt( (x2-x1)^2 + (y2-y1)^2 ))
	return math.sqrt( (x2-x1)^2 + (y2-y1)^2 )
end

-- Actor->Position Pythagorean Distance
function APDistance(actor,x,y)
	-- Check actor
	if type(actor) ~= "number" then
		-- Return nil so its easier to trace
		return nil
	end

	local x2,y2 = GetV(V_POSITION,actor)

	if x2 ~= -1 then
		return PPDistance(x2,y2,x,y)
	end

	-- If there was an invalid actor, return a high number
	return SightRange * SightRange
end

-- Actor->Actor Pythagorean Distance
function AADistance(actor1,actor2)
	-- Check actor2
	if type(actor2) ~= "number" then
		-- Return nil so its easier to trace
		return nil
	end

	local x,y = GetV(V_POSITION,actor2)

	if x ~= -1 then
		return APDistance(actor1,x,y)
	end

	-- If there was an invalid actor, return a high number
	return SightRange * SightRange
end

-- Position->Position Block Distance
function PPBlocks(x1,y1,x2,y2)
	local xdist = math.abs(x2-x1)
	local ydist = math.abs(y2-y1)

	if xdist > ydist then
		return xdist
	else
		return ydist
	end
end

-- Actor->Position Block Distance
function APBlocks(actor,x,y)
	-- Check actor
	if type(actor) ~= "number" then
		-- Return nil so its easier to trace
		return nil
	end

	local x2,y2 = GetV(V_POSITION,actor)

	if x2 ~= -1 then
		return PPBlocks(x2,y2,x,y)
	end

	-- If there was an invalid actor, return a high number
	return SightRange * SightRange
end

-- Actor->Actor Block Distance
function AABlocks(actor1,actor2)
	-- Check actor 2
	if type(actor2) ~= "number" then
		-- Return nil so its easier to trace
		return nil
	end

	local x,y = GetV(V_POSITION,actor2)

	if x ~= -1 then
		return APBlocks(actor1,x,y)
	end

	-- If there was an invalid actor, return a high number
	return SightRange * SightRange
end

-- Find actor nearest to a position from a table
function PTNearest(x,y,t)
	-- Check the size of the table
	if Table.getn(t) < 1 then
		return -1
	end

	-- Loop through the table
	local i,dist=0,SightRange * SightRange
	local nearest=dist
	local index=-1
	for i=1,t.n do
		-- Make sure its an actor number
		if type(t[i]) == "number" then
			-- Get the distance to the actor
			dist=APDistance(t[i],x,y)
		end

		-- Check if its closer
		if dist < nearest then
			-- Set the new nearest actor
			nearest=dist
			index=i
		end
	end

	-- Return the nearest
	return t[index]
end

-- Find actor nearest to another from a table
function ATNearest(actor,t)
	-- Check the size of the table
	if Table.getn(t) < 1 then
		return -1
	end

	-- Get the actor's current position
	local x,y = GetV(V_POSITION,actor)

	-- Get the nearest from the table
	return PTNearest(x,y,t)
end

---------------------------
-- Sight-range Functions --
---------------------------

-- Check if a position is in sight-range of our owner
function IsInSight(x,y)
	if APBlocks(OwnerID,x,y) > SightRange then
		return false
	end

	return true
end

-- Check if an actor is in sight-range of our owner
function ActorIsInSight(actor)
	if AABlocks(actor,OwnerID) > SightRange then
		return false
	end

	return true
end

----------------------
-- Circle Functions --
----------------------

-- Function to get the angle from one point to another
function PPGetAngle(x1,y1,x2,y2)
	-- Get the radius
	local radius = PPDistance(x1,y1,x2,y2)

	-- Check if the radius is 0
	if radius == 0 then
		-- Return an angle of -1
		return -1,0
	end

	-- Get the changes
	local xdelta = x2 - x1
	local ydelta = y2 - y1

	-- Get the angle from X
	local angle = math.deg(math.acos(xdelta / radius))

	-- Get the angle from Y
	if math.deg(math.asin(ydelta / radius)) < 0 then
		angle = 360 - angle
	end

	-- Return the angle and the radius
	return angle,radius
end

-- Function to get the angle from a point to an actor
function PAGetAngle(x1,y1,actor)
	-- Get the actor's coordinates
	local x2,y2 = GetV(V_POSITION,actor)

	-- Check the actor's coordinates for sanity
	if x2 == -1 then
		-- Actor's position is unknown
		return -1,0
	end

	-- Reuse above code
	return PPGetAngle(x1,y1,x2,y2)
end

-- Function to get the angle from an actor to a position
function APGetAngle(actor,x2,y2)
	-- Get the actor's coordinates
	local x1,y1 = GetV(V_POSITION,actor)

	-- Check the actor's coordinates for sanity
	if x1 == -1 then
		-- Actor's position is unknown
		return -1,0
	end

	-- Reuse above code
	return PPGetAngle(x1,y1,x2,y2)
end

-- Function to get the angle from one actor to another
function AAGetAngle(actor1,actor2)
	-- Get the first actor's coordinates
	local x,y = GetV(V_POSITION,actor1)

	-- Check the actor's coordinates for sanity
	if x == -1 then
		-- Actor1's position is unknown
		return -1,0
	end

	-- Reuse above code
	return PAGetAngle(x,y,actor2)
end

-- Function to plot a point on a circle centered around a position
function PPlotCircle(x,y,angle,radius)
	-- Get the angle in radians
	local rads = math.rad(angle)

	-- Get the xdelta and ydeltas
	local xdelta = radius * math.cos(rads)
	local ydelta = radius * math.sin(rads)

	-- Return the locations
	return x + xdelta, y + ydelta
end

-- Function to plot a point on a circle centered around an actor
function APlotCircle(actor,angle,radius)
	-- Get the actor's coordinates
	local x,y = GetV(V_POSITION,actor)

	-- Check the actor's coordinates for sanity
	if x == -1 then
		-- Return -1s for the coordinates
		return -1,-1
	end

	-- Reuse above code
	PPlotCircle(x,y,angle,radius)
end

-- Check to see if an angle is within X radians of a second
function AngleCompare(angle1,angle2,radians)
	local bottom,top

	-- Get the modulus remainder of both angles
	angle1 = math.mod(angle1,360)
	angle2 = math.mod(angle2,360)

	-- Check if radians is positive or negative
	if radians > 0 then
		-- Check if adding radians to angle1 would throw it past 359
		if angle1 + radians >= 360 then
			-- Make a second check
			if AngleCompare(0,angle2,math.mod(angle1 + radians,360)) then
				-- Found it already
				return true
			end

			-- Reduce radians below 360
			radians = radians - math.mod(angle1 + radians,360) - 1
		end

		-- Set the bottom and the top
		bottom = angle1
		top = angle1 + radians

	elseif radians < 0 then
		-- Invert the radians, to make later stuff easier to read
		radians = -radians

		-- Check if subtracting radians would throw it under 0
		if angle1 - radians < 0 then
			-- Make a second check
			if AngleCompare(360 + angle1 - radians,angle2,math.abs(angle1 - radians) - 1) then
				-- Found it already
				return true
			end

			-- Increase the radians above -1
			radians = radians + math.abs(angle1 - radians)
		end

		-- Set the bottom
		bottom = angle1 - radians
		top = angle1

	else
		-- If radians is 0, then check if the angles are the same
		if angle1 == angle2 then
			-- They check out as the same
			return true
		end

		-- They're not the same
		return false
	end

	-- Check if its above the bottom but below the top, inclusively
	if bottom <= angle2 and angle2 <= top then
		-- Its inside
		return true
	end

	-- Its not inside
	return false
end

-------------------
-- Actor Support --
-------------------

-- Check if an actor is in a table
function ActorIsInTable(actor,t)
	-- Loop through the table
	local i,max = 1,Table.getn(t)
	for i=1,max do
		-- Check if its the actor
		if t[i] == actor then
			-- It is, return true
			return true
		end
	end
	
	-- Otherwise the actor isn't in the table
	return false
end

-- Identify various bits of information, such as all the attackers
-- of a specific actor, members of the party, etc
function GetActorInfo()
	-- Get all the actors on the screen
	local actors = GetActors()

	-- Make a new table of actor info
	local newData = { }

	-- Clear our table of temporary friends
	TempFriends = { }

	-- Clear our table of permanent friends that are on screen
	PermFriendsOnScr = { }

	-- Get the current time
	local ticks = GetTick()

	-- Get the owner's location
	local ox,oy = GetV(V_POSITION,OwnerID)

	-- Temporary function to create a new entry
	local newEntry = function(actor,type)
		-- Check if the actor is dead
		if ActorData[actor] and ticks - ActorData[actor].First.Tick > 2000 and
			GetV(V_MOTION,actor) == MOTION_DEAD then
			-- Remove old data, so its fresh
			ActorData[actor] = nil
		end

		-- Check if we have data from before
		if ActorData[actor] ~= nil then
			-- Move it to the new table
			local ret = ActorData[actor]
			ActorData[actor] = nil
			return ret
		end

		-- Make sure we have a type
		if not type then
			-- Get the type
			type = GetV(V_HOMUNTYPE,actor)

			-- Remove overlap of homunculi and player jobs
			if actor <= 99999 and LIF <= type and type <= VANILMIRTH_H2 then
				-- Its a homunculus, so add 6000 to its ID
				type = type + 6000
			end
		end

		-- Log that its first time seeing this actor
		Log(9,"First sight of actor ID #%d (type %d)",actor,type)

		-- Create new data since they didn't have previous data
		local x,y = GetV(V_POSITION,actor)
		local ret = {
			First = { Tick=ticks, X=x, Y=y, OwnerX=ox, OwnerY=oy },
			Last = {
				-- Previous location
				X=x, Y=x,
				PosChange=ticks,	-- time position changed

				-- Motion last cycle
				Motion=0,

				-- The tick time of the last time the motion changed
				MotionChange=0,

				-- Ticks of the last motions
				Attack=0,
				Move=0,
				Damage=0,
				Cast=0,
				Skill=0,
			},
			Count = {
				-- Number of times they've been auto-healed
				Heal=0,

				-- Number of times these motions have been seen
				--Attack=0
				Damage=0,
				--Cast=0,		not needed/used
				--Skill=0,

				-- Number of times they've been casted against
				HomuSkill=0,
				AlchSkill=0,
				AlchMammo=0,
				AlchAcidT=0,
				AlchAcidB=0,
			},
			X=x, Y=x,
			Type = type,
			Motion = 0,
			Target = 0,
			TargetOf = { update = 0, },

			-- Actor's "partner" (aka: a homunculus's owner, or an alchemist's homu)
			Partner = 0,
			NeverPartner = false,

			-- Should we check to see if they're invulnerable for 3 seconds?
			PossibleInvuln = true,

			-- Possible actors that this one is chasing
			ChaseTargets = Table.new(),
		}

		-- Check if we want to search for chase-targets
		if false and PPBlocks(x,y,ox,oy) >= 14 and RAI_IsMonster(actor) == 1 then
			-- Loop through the old actors, finding one that appeared near this one
			local actor,t
			for actor,t in pairs(ActorData) do
				-- Check if they appeared at the edge of the screen, aren't a monster,
				--	aren't a friend, and appeared nearby where the return monster appeared
				if PPBlocks(t.First.OwnerX,t.First.OwnerY,t.First.X,t.First.Y) >= 14 and
					RAI_IsMonster(actor) == 0 and not IsFriend(actor) and
					PPDistance(x,y,t.First.X,t.First.Y) < 3 then

					-- Add them to the potential chase targets
					Table.insert(ret.ChaseTargets,actor)
				end
			end
		end

		return ret
	end

	-- Clear the owner and homunculus's TargetOf table
	HomunTargetOf = { update=ticks, }
	OwnerTargetOf = { update=ticks, }

	-- Temporary function to add to an actor's TargetOf table
	local newTargetOf = function(actor,target)
		-- Check if the actor is targetting anything
		if target ~= 0 then
			local targetOfTable
			if target == HomunID then
				targetOfTable = HomunTargetOf
			elseif target == OwnerID then
				targetOfTable = OwnerTargetOf
			else
				-- Check if the target has a table yet
				if newData[target] == nil then
					-- Check that the target's type isn't -1
					if GetV(V_HOMUNTYPE,target) == -1 then
						return
					end

					-- Nope, create a table for it
					newData[target] = newEntry(target)
				end

				-- Check to see if the TargetOf is recent
				if newData[target].TargetOf.update < ticks then
					-- Make a new TargetOf table
					newData[target].TargetOf = { update=ticks, }
				end

				-- Get the table
				targetOfTable = newData[target].TargetOf
			end

			-- Add the actor as an attacker of the target
			Table.insert(targetOfTable,actor)
		end
	end

	-- Loop through each actor, gathering data
	local i,actor,target,nearestPortal = nil,nil,nil,100
	local unmatchedHomus,unmatchedAlchs = {},{}
	for i,actor in ipairs(actors) do
		-- Make sure its a valid actor
		if actor ~= -1 then
			-- Get the type
			local type = GetV(V_HOMUNTYPE,actor)

			-- Remove overlap of homunculi and player jobs
			if actor <= 99999 and
				LIF <= type and type <= VANILMIRTH_H2 then
				-- Its a homunculus, so add 6000 to its ID
				type = type + 6000
			end

			-- Don't gather data on HomunID or OwnerID, as it'll be redundant
			if actor == HomunID or actor == OwnerID then

				-- Add their target to the TargetOf table
				-- newTargetOf(actor,GetV(V_TARGET,actor))

				-- NOTE: But theres no point.

			-- Make sure we don't gather data on a portal
			elseif type == 45 then

				-- Check the distance of the portal
				local blocks = APBlocks(actor,ox,oy)
				if blocks < nearestPortal then
					-- This portal is closer
					nearestPortal = blocks
				end

			-- Don't gather data on an NPC
			elseif type < 1000 and
				RAI_IsMonster(actor) ~= 1 and

				-- The following are indicators of player jobs, but its
				--	negated with "not" because player jobs are not NPCs
				not ((0 <= type and type <= 25) or
				(161 <= type and type <= 181) or
				(4001 <= type and type <= 4049)) then

				-- Do nothing, its probably an NPC

			-- Everything else we should gather data on
			elseif type ~= -1 then

				-- Check if we already have data
				if newData[actor] == nil then
					newData[actor] = newEntry(actor,type)
				end
				local a = newData[actor]

				-- Check if the type changed
				if type ~= a.Type then
					Log(2,"Actor #%d changed type from %d to %d.",actor,a.Type,type)
					a.Type = type
				end

				-- Check if the location changed
				local x,y = GetV(V_POSITION,actor)
				if x ~= a.X or y ~= a.Y then
					a.Last.X,a.Last.Y = a.X,a.Y
					a.Last.PosChange = ticks
					a.X,a.Y = x,y
				end

				-- Get the actor's motion
				a.Last.Motion = a.Motion
				a.Motion = GetV(V_MOTION,actor)

				-- Check if the motion changed
				if a.Motion ~= a.Last.Motion then
					-- Set the time that the motion changed
					a.Last.MotionChange = ticks
				end

				-- Check to see if the TargetOf list got updated this tick
				if a.TargetOf.update < ticks then
					-- Clear the TargetOf list
					a.TargetOf = { update=ticks, }
				end

				-- Check if the actor is attacking
				if a.Motion == MOTION_ATTACK or a.Motion == MOTION_ATTACK2 then
					-- They're attacking, so set the last attack time to now
					a.Last.Attack = ticks

				-- Check if the actor is moving
				elseif a.Motion == MOTION_MOVE then
					-- Set the move time to right now
					a.Last.Move = ticks

				-- Check if the actor is getting damaged
				elseif a.Motion == MOTION_DAMAGE then
					-- They're getting damaged, so set their
					--	damaged time to right now
					a.Last.Damage = ticks

					-- Update the number of times they've been damaged
					if a.Last.Motion ~= a.Motion or
						ticks > a.Last.MotionChange + 500 then

						-- Increment the damage count
						a.Count.Damage = a.Count.Damage + 1

						-- Reset motion change, so we only increment
						--	once so often if the actor is
						--	constantly in "damaged" state
						a.Last.MotionChange = ticks
					end
		
				-- Check if they're casting
				elseif a.Motion == MOTION_CASTING then
					-- Set the casting time to right now
					a.Last.Cast = ticks

				-- Check if they're using a skill
				elseif a.Motion == MOTION_SKILL or
					a.Motion == MOTION_TOSS or
					a.Motion == MOTION_BIGTOSS then
					-- Set the skill time to right now
					a.Last.Skill = ticks
		
				end

				-- Check if the actor is in a position to target something
				if motion ~= MOTION_DEAD and motion ~= MOTION_SIT then
					-- It is
		
					-- Get the current actor's target
					a.Target = GetV(V_TARGET,actor)
		
					-- Add it to the TargetOf table
					newTargetOf(actor,a.Target)
				end

				-- Check if the actor is a permanent friend
				if IsPermFriend(actor) then
					-- Add it to our friends who're on screen
					PermFriendsOnScr[actor] = true
				end

				-- Should we gather friend data?
				if DefendFriends and actor ~= HomunID and actor ~= OwnerID then
					-- Check if the actor is a monster or a homunculus
					if RAI_IsMonster(actor) ~= 1 and
						(6000 >= type or type >= 6017) then
						-- Its neither, so allow it to be a temporary friend

						-- Check if they're close enough to our owner
						if AABlocks(OwnerID,actor) <= TemporaryFriendTiles then
							-- They're a temporary friend
							TempFriends[actor] = true
						end
					end

					-- If we're gathering friend data, then there really is a
					--	reason to know which homunculi are matched to which
					--	alchemists
					if not a.NeverPartner then
						if a.Partner == 0 then
							-- Alchemist / Biochemist type 1 / Biochemist type 2 / Baby Alchemist
							if type == 18 or type == 179 or type == 4019 or type == 4041 then
								-- unmatched alchemist
								unmatchedAlchs[actor] = true
							elseif 6001 <= type and type <= 6016 then
								-- unmatched homunculus
								unmatchedHomus[actor] = true
							else
								-- Never check this actor again (it can't have a partner)
								a.NeverPartner = true
							end
						end
					end
				end

			end
		end	-- actor ~= -1
	end

	-- Check if the actor disappeared
	for actor,i in pairs(ActorData) do
		-- Check if they had a partner
		if not i.NeverPartner and i.Partner ~= 0 and
			newData[i.Partner] ~= nil then
			-- The partner is now solo
			newData[i.Partner].Partner = 0

			-- Decide if its a homunculus, and add it to the unmatched table
			if i.Partner < 100000 then
				-- homunculus
				unmatchedHomus[i.Partner] = true
			else
				-- alchemist
				unmatchedAlchs[i.Partner] = true
			end
		end

		-- Log that it disappeared
		Log(3,"Actor ID #%d (type %d) disappeared.",actor,i.Type)

		-- Remove it, make it easier on garbage collector
		ActorData[actor] = nil
	end
	i = nil

	-- Find matches for each of the alchemists who don't have a partner
	local j,alchT,homuT,nearEdge,matched,dummy
	for i,dummy in pairs(unmatchedAlchs) do
		if dummy then
			-- Not matched yet
			matched = false
	
			-- Get the actor data table
			alchT = newData[i]
	
			-- Check if the alchemist is near the edge of the screen
			if PPBlocks(alchT.X,alchT.Y,ox,oy) < 14 then
				nearEdge = false
			else
				nearEdge = true
			end
	
			-- Loop through the homunculi
			local tenativeMatch = 0
			local tenativeDist = 14
			for j,dummy in pairs(unmatchedHomus) do
				-- Get the data table
				homuT = newData[j]

				-- Get the block distance between homun and alchie
				local blocks = PPBlocks(homuT.X,homuT.Y,alchT.X,alchT.Y)
	
				-- Make sure the homunculus and the alchemist are close enough
				if blocks <= 14 then
	
					-- Check if the alchemist isn't at the edge
					if not nearEdge and
	
					-- Check if the homunculus spawned underneath the alch
						(homuT.First.Ticks == ticks and blocks < 1) then
	
						-- Its a match
						homuT.Partner = i
						alchT.Partner = j

						-- Remove the unmatched homu
						unmatchedHomus[j] = nil
		
						-- They're matched
						matched = true
						tenativeMatch = 0
		
						-- Stop looping through the homus
						break
					end
		
					-- Poor decision, but if its the closest, then tenatively set it as match
					if blocks < tenativeDist then
						tenativeMatch = j
						tenativeDist = blocks
					end
	
				end
			end
	
			-- Check if there was a tenative match
			if tenativeMatch ~= 0 then
				homuT.Partner = i
				alchT.Partner = tenativeMatch
	
				-- Remove the unmatched homu
				unmatchedHomus[tenativeMatch] = nil
	
				-- They're matched
				matched = true
			end
	
			-- Check if this alchemist got matched
			if matched and DebugLevel > 2 then
				-- Log it
				Log(3,"Partner-matching: actor #%d matched to actor #%d.",
					i,alchT.Partner)
	
				-- Remove it if we're going to tell the unmatched alchs/homus
				unmatchedAlchs[i] = false
			end
		end
	end

	-- Set the new data as the ActorData
	ActorData = newData

	-- Check the distance of the nearest portal
	local entPort = false
	if (nearestPortal <= 4 and nearestPortal < OwnerLast.PortalDist
		and OwnerMotion == MOTION_MOVE) or
		nearestPortal <= 1 then

		-- We're probably entering a portal
		entPort = true
	end
	OwnerLast.PortalDist = nearestPortal

	-- Check if we're entering a portal
	if entPort and not EnteringPortal then
		-- Log it
		Log(1,"Potentially entering a portal, disabling all script " ..
			"actions temporarily.")

		-- Switch it on
		EnteringPortal = true
	elseif not entPort and EnteringPortal then
		-- Log it
		Log(1,"Re-enabling script actions, portal entrance was seemingly false-alarm.")

		-- Switch it off
		EnteringPortal = false
	end
end

-------------------
-- HP/SP Support --
-------------------

-- Calculate alchemist's HP delta
function AI_CalculateHPDeltasAlch()
	-- Get the current tick count
	local ticks = GetTick()

	-- Get our owner's HP value
	local hp = GetV(V_HP,OwnerID)

	-- Remove any HP values that are older than 1000 milliseconds
	local other
	repeat
		-- Grab the oldest HP value
		other = List.popright(OwnerHPTracking)

		-- Check the validity
		if other == nil then
			other = { ticks, hp }
		end
	until ticks - other[1] <= 1000

	-- Its new enough, push it back on
	List.pushright(OwnerHPTracking,other)

	-- Get the delta
	OwnerHPDelta.second = hp - other[2]

	-- Now grab the newest HP value
	other = OwnerHPTracking[OwnerHPTracking.first]

	-- Get the delta
	OwnerHPDelta.iteration = hp - other[2]

	-- Now add this into our log
	List.pushleft(OwnerHPTracking,{ ticks, hp })
end

-- Calculate homunculus's HP delta
function AI_CalculateHPDeltasHomu()
	-- Get the current tick count
	local ticks = GetTick()

	-- Begin doing the exact same for homunculus
	local hp = GetV(V_HP,HomunID)

	-- Remove any HP values that are older than 1000 milliseconds
	local other
	repeat
		-- Grab the oldest HP value
		other = List.popright(HomunHPTracking)

		-- Check the validity
		if other == nil then
			other = { ticks, hp }
		end
	until ticks - other[1] <= 1000

	-- Its new enough, so put it back
	List.pushright(HomunHPTracking,other)

	-- Get the delta
	HomunHPDelta.second = hp - other[2]

	-- Now grab the newest HP value
	other = HomunHPTracking[HomunHPTracking.first]

	-- Get the delta
	HomunHPDelta.iteration = hp - other[2]

	-- Now add this into our hp log
	List.pushleft(HomunHPTracking,{ ticks, hp })
end

-- Determine the amount of HP remaining
function ActorHPAmount(actor)
	-- Get the actor's HP amount
	return GetV(V_HP,actor)
end

-- Determine the percentage of SP remaining
function ActorHPPercent(actor)
	-- Get the actor's HP and max HP amounts
	local hp = GetV(V_HP,actor)
	local max = GetV(V_MAXHP,actor)

	-- Turn it into a percentage and return
	return math.floor( (hp/max) * 100 )
end

-- Determine the maximum amount of SP that is usable
function SPUsableAmount(actor)
	-- Decide on the minimum amount
	local min = 0

	-- Check if we're getting our homunculus's SP or our alchemist's SP
	if actor == HomunID then
		min = MinSP
	elseif actor == OwnerID then
		min = OwnerMinSP

		-- Check if we have autores enabled
		if AutoRes then
			min = min + 50
		end
	end

	-- Subtract our minimum SP amount
	return ActorSP(actor) - min
end

-- Determine the maximum amount of SP that is usable, using percents
function SPUsablePercent(actor)
	-- Get the percent our SP is at
	local sp = ActorSPPercent(actor)

	-- Decide on the minimum amount
	local min = 0

	-- Check if we're getting our homunculus's SP or our alchemist's SP
	if actor == HomunID then
		min = MinSP
	elseif actor == OwnerID then
		min = OwnerMinSP
	end

	-- Determine the difference
	sp = sp - min

	-- Unpercentize it
	sp = math.floor((sp * GetV(V_MAXSP,actor) / 100))

	-- Check if we need to save some for autores
	if actor == OwnerID and AutoRes then
		-- Save 50 for AutoRes
		sp = sp - 50
	end

	-- Return the SP
	return sp
end

-- Determine the amount of SP remaining
function ActorSPAmount(actor)
	-- Get the actor's SP amount
	return GetV(V_SP,actor)
end

-- Determine the percentage of SP remaining
function ActorSPPercent(actor)
	-- Get the actor's SP and max SP amounts
	local sp = GetV(V_SP,actor)
	local max = GetV(V_MAXSP,actor)

	-- Turn it into a percentage and return
	return math.floor( (sp/max) * 100 )
end

-----------------------------
-- String Buffer Functions --
-----------------------------

-- (idea from http://www.lua.org/notes/ltn009.html)

StringBuffer = {}

-- Create a new string buffer
StringBuffer.New = function()
	return List.new()
end

-- Append data
StringBuffer.Append = function(buf,data)
	-- Append the data
	List.pushright(buf,data)

	-- Now clean it up
	local i
	for i=buf.last-1,buf.first,-1 do
		-- Check if the string is shorter than the previous
		if string.len(buf[i+1]) > string.len(buf[i]) then break end

		-- Concatenate the strings
		buf[i] = buf[i] .. List.popright(buf)
	end
end

-- Get the string
StringBuffer.Get = function(buf)
	-- Concatenate the entire thing into 1 string
	local i
	for i=buf.last-1,buf.first,-1 do
		-- Concatenate the strings
		buf[i] = buf[i] .. List.popright(buf)
	end

	-- Return the string
	return buf[buf.first]
end

-------------------------
-- Debugging Functions --
-------------------------

-- Normal log-style
function Log(level,text,...)
	-- Check the level is too high
	if level > DebugLevel then
		return
	end

	-- Format the string
	local str = "(" .. level .. ") " .. LogFormat(text,unpack(arg))

	-- Check if its the same as the last log line
	if LastLogLine == str then
		-- Add a flag to say its a duplicate
		str = str .. " (duplicate)"
	end

	-- Save the log line
	LastLogLine = str

	-- Make the actual log
	RAI_TraceAI(str)
end

-- Workaround for logs when TraceAI.txt is only 1 line
function LogWorkaround(level,text,...)
	-- Check if the level is too high
	if level > DebugLevel then
		return
	end

	-- Format the text
	local str = "(" .. level .. ") " .. LogFormat(text,unpack(arg))

	-- Put the text into the log
	List.pushright(Logs,str)

	-- See if we need to remove a line
	if List.size(Logs) > LogLines then
		List.popleft(Logs)
	end

	-- Gather all the lines into 1 string
	str = StringBuffer.New()
	StringBuffer.Append(str,"\r\n")
	local i
	for i=Logs.first,Logs.last do
		StringBuffer.Append(str,Logs[i])
		StringBuffer.Append(str,"\r\n")
	end

	-- Output via TraceAI
	RAI_TraceAI(StringBuffer.Get(str))
end

-- Function to verify that we have enough parameters to a string.format, and then format it
function LogFormat(text,...)
	-- Default to having valid parameters
	local valid = true

	-- Find all places that a % operator shows up
	local curParam,paramStart,paramEnd = 1,nil,1
	while true do
		-- Get the next match
		paramStart,paramEnd = string.find(text,"%%%d-%.?%d-[dfiqs]",paramEnd)

		-- Check that we found a place
		if paramStart == nil then
			-- We didn't, stop looping
			break
		end

		-- Get the type that we need
		local t = string.sub(text,paramEnd,paramEnd)
		if t == "q" or t == "s" then
			-- We need a string
			t = "string"
		else
			-- We need a number
			t = "number"
		end

		-- Check if our parameter is of that type
		if type(arg[curParam]) ~= t then
			-- Its not, set valid to false
			valid = false

			-- Stop looping
			break
		end

		-- Increment our current parameter
		curParam = curParam + 1
	end

	-- Check if we're valid
	if valid then
		-- Format as normal
		return string.format(text,unpack(arg))
	else
		-- Create a string buffer, for many concatenations
		local str = StringBuffer.New()

		-- Format to an error
		StringBuffer.Append(str,string.format("Error with LogFormat(%q",text))

		-- Add all the arguments
		local i,form
		for i=1,Table.getn(arg) do
			-- Check if we should put quotes around it
			if type(arg[i]) == "string" then
				form = "%q"
			else
				form = "%s"
			end
			-- Add the parameter
			StringBuffer.Append(str,string.format(","..form,tostring(arg[i])))
		end

		-- Add the end parenthesis
		StringBuffer.Append(str,")")

		-- Return our string
		return StringBuffer.Get(str)
	end
end

---------------------------
-- Function Replacements --
---------------------------

function NoReturn()
end

function ZeroReturn()
	return 0
end

function ZeroZeroReturn()
	return 0,0
end

function NegativeReturn()
	return -1
end

function TrueReturn()
	return true
end

function FalseReturn()
	return false
end

function PassThru(arg)
	return arg
end